% *** ARD-margx Algorithm ***
% *algorithm based on Wipf's and Nagarajan's approach using analytical
%  marginalization over vector x
%  to solve set of equations m = Ax + e, where x is sparse vector of unknowns
% *see paper 'Comparison of sparse Bayesian regularization methods
%  for high-dimensional inverse problem'
% *'visualisation of the reconstruction' is prepared for case,
%  where x is n*n image - a modification is easy and possible
%
% % written by Jakub Sevcik and Vaclav Smidl 2016

% Load the measurement matrix (A) and the vector (m)
% *a training dataset is downloadable from http://www.fips.fi/dataset.php
load Data328 A m
m = m(:);

% % ARD-margx Algorithm
% inicialization
iterations = 10;
iterationsNLCG = 20; % number of iteration that use NonLinear Conjugate Gradients method

[M,N] = size(A);
n = sqrt(N);

beta = 1;
z = ones(N,1);

% calculate Ap2 = (A.^2)' for faster computation
Ap2 = sparse(A.^2);

% iterative section
for iter = 1:1:iterations
    NAtb = norm(A'*m);
    fun = @(x) (A*x - m)'*(A*x - m) + 2/beta*(sqrt(z)'*abs(x));
    grad = @(x) 2*A'*(A*x - m) + 2/beta*sqrt(z).*sign(x);
    
    x = nlcgClever(fun,grad,N,NAtb,iterationsNLCG);   
    
    alfaInv = abs(x)./sqrt(z);  
    
    dSb = Ap2*alfaInv+1/beta*ones(M,1);
    idSb = 1./dSb;
    z = Ap2'*idSb;
    
    if mod(iter,10) == 0
        disp([iter, iterations])
    end
end

funw   = @(w) A*(alfaInv.*(A'*w))+1/beta*w;
w = pcg(funw,m,[],20);
x = alfaInv.*(A'*w);

% % visualisation of the reconstruction
figure
imagesc(reshape(x,n,n))
colormap gray
axis square
axis off
title(['ARD-margx Algorithm with ',num2str(iterations),' iterations'])
